/*
 * Control of the balancing robot Balboa 32U4 
 * as a part of the research by XY.
 * This implementation is done by XY in September 2023.
 * 
 */
// general libaries to load 
#include <Balboa32U4.h> // preprogrammed functions of the Balboa 32U4
#include <LSM6.h> // data processing of the IMU (used hardware component)
#include <MadgwickAHRS.h>

// declare general variables
// adapt factors to normalize the output of the IMU 
float adapt_factor_gyro = 28.57;
float adapt_factor_accelerometer = 8196.7213;
float adapt_factor_encoder = 1779.6;

// initialize state space memorys
float gyro_offsets[3]; // for each coordinate direction x, y, z
float imu_measurement[4]; // [roll, pitch, heading, pitch_processed] 
float encoder_measurement[10]; // [currentTime, previousTime, countsLeft, countsRight, phiLeft, phiRight, dPhiLeft, dPhiRight]
float state[4]; // [theta, phi, dtheta, dphi]

// Firstly, the inertial measurement unit (IMU) and the encoder 
// for state representation are prepared
// setup the inertial measurement unit 
LSM6 imu;
// setup the encoder 
Balboa32U4Encoders encoders;
void initializeEncoder(){
  encoder_measurement[0] = millis();
  encoder_measurement[2] = encoders.getCountsAndResetLeft();
  encoder_measurement[3] = encoders.getCountsAndResetRight();
  encoder_measurement[4] = calculate_Phi(encoder_measurement[2] ,adapt_factor_encoder, true);
  encoder_measurement[5] = calculate_Phi(encoder_measurement[3] ,adapt_factor_encoder, true);
}
// setup the Madgwick filter
Madgwick filter; 

// The angle is calulated using data from the accelerometer and gyroscope
// The Madgwick toolbox was used for this

void initializeMadgwickFilter() {
  initializeIMU();
  const float sensorRate = 208.00;
  filter.begin(sensorRate);
}
// helperfunction for offset treatment
void initializeIMU(){

  // gyro offset treatment
  float gyro_offset_x = 0; 
  float gyro_offset_y = 0; 
  float gyro_offset_z = 0; 
  
  // write the accelerometer settings 
  imu.writeReg(LSM6::CTRL1_XL, 0b01011000); // 208 Hz, +/- 4 m/s^2
  // write the gyro settings
  imu.writeReg(LSM6::CTRL2_G, 0b01011000); // 208 Hz, 1000 deg/s

  // wait for IMU readings to stabilze
  delay(1000);
  
  // calibrate the gyro 
  for (int i=0; i<1000; i++){
      imu.read();
      gyro_offset_x+=imu.g.x;
      gyro_offset_y+=imu.g.y;
      gyro_offset_z+=imu.g.z;
    }

  gyro_offset_x = gyro_offset_x/1000;
  gyro_offset_y = gyro_offset_y/1000;
  gyro_offset_z = gyro_offset_z/1000;

  // return the calibration subtrahend
  float gyro_offsets[3];
  gyro_offsets[0] = gyro_offset_x;
  gyro_offsets[1] = gyro_offset_y;
  gyro_offsets[2] = gyro_offset_z;
}

// evaluation of the inertial measurement unit (calculation of theta and dtheta)
void evaluateIMU(){
  imu.read();
  // normalize the sensor raw data
  float* gyroNormalized = normalizeGyro(imu.g.x, imu.g.y, imu.g.z, adapt_factor_gyro, gyro_offsets);
  float* accNormalized = normalizeAcc(imu.a.x, imu.a.y, imu.a.z, adapt_factor_accelerometer); 
  // run the Madgwick filter
  filter.updateIMU(gyroNormalized[0], gyroNormalized[1], gyroNormalized[2], accNormalized[0], accNormalized[1], accNormalized[2]);
  // restore the filter results
  imu_measurement[0] = filter.getRoll(); 
  imu_measurement[1] = filter.getPitch(); 
  imu_measurement[2] = filter.getYaw(); 
  imu_measurement[3] = pitch_dataprocessing(imu_measurement[1], imu_measurement[0]); // calculate the defined rest point in the state space

  // save as a part of the state space vector
  state[0] = imu_measurement[3]; // theta
  state[2] = gyroNormalized[1];  // dtheta 
}

float* normalizeGyro(float gyroX, float gyroY, float gyroZ, float adapt_factor, float offsets[3]){
  float gyroNormalized[3]; 
  gyroNormalized[0] = ((float) gyroX - offsets[0])/adapt_factor;
  gyroNormalized[1] = ((float) gyroY - offsets[1])/adapt_factor;
  gyroNormalized[2] = ((float) gyroZ - offsets[2])/adapt_factor;
  return gyroNormalized;
}

float* normalizeAcc(float accX, float accY, float accZ, float adapt_factor){
  float accNormalized[3]; 
  accNormalized[0] = accX/adapt_factor; 
  accNormalized[1] = accY/adapt_factor; 
  accNormalized[2] = accZ/adapt_factor;
  return accNormalized;  
}

float pitch_dataprocessing(float pitch, float roll){
  if(abs(roll)>=90){
    return (pitch + 90); 
  }else{
    return abs(pitch)-90;
  }
}

// evaluation of the Encoders (for balance control: left and right are equal) 
void evaluateEncoders(){
  encoder_measurement[1] = encoder_measurement[0]; // previousTime becomes currentTime
  encoder_measurement[0] = millis(); // set new currentTime

  // save the current number of counts (output of the encoder)
  encoder_measurement[2] = encoders.getCountsAndResetLeft();
  encoder_measurement[3] = encoders.getCountsAndResetRight();
  
  // save the change of the angle
  float curr_dPhiLeft = calculate_Phi(encoder_measurement[2], adapt_factor_encoder, true);
  float curr_dPhiRight = calculate_Phi(encoder_measurement[3], adapt_factor_encoder, true);

  // accumulate the actual phi
  encoder_measurement[4] += curr_dPhiLeft;
  encoder_measurement[5] += curr_dPhiRight;

  // calculate the speed of the wheel rotation (radian per second) 
  encoder_measurement[6] = calculate_dPhi(curr_dPhiLeft, encoder_measurement[0], encoder_measurement[1]);
  encoder_measurement[7] = calculate_dPhi(curr_dPhiRight, encoder_measurement[0], encoder_measurement[1]);

  // save as a part of the state space vector
  state[1] = encoder_measurement[4]; // or encoder_measurement[5] (equals for the balance regulator, because the torque is applied equally)
  state[3] = encoder_measurement[6]; // or encoder_measurement[7] (equals for the balance regulator, because the torque is applied equally)
}

// calulate phi under consideration of the gearbox and 12 time counts per revolution
float calculate_Phi(float counts, float adapt_factor_encoder, bool radian){
  if (radian==true){
    return (2*PI*counts)/adapt_factor_encoder;
  }else{
    return (360*counts)/adapt_factor_encoder;
  }
}
// calulate dphi under consideration that currentTime and previousTime is measured in milliseconds
float calculate_dPhi(float current_dPhi, float currentTime, float previousTime){
  return (current_dPhi)/(currentTime-previousTime);
}


// run the programm
void setup() {
  // initialize Madgwick filter including the IMU 
  initializeMadgwickFilter(); 
  initializeEncoder();
}

void loop() {
  // update theta and dtheta (saved as state[0] and state[2])
  evaluateIMU();
  // update phi and dphi (saved as state[1] and state[3])
  evaluateEncoders();
  
  // prints to check the results
  Serial.print("State:   theta:"); 
  Serial.print(state[0]); 
  Serial.print("   phi:"); 
  Serial.print(state[1]);
  Serial.print("   dtheta:"); 
  Serial.print(state[2]); 
  Serial.print("   dphi:"); 
  Serial.println(state[3]);
}
